@@ -19,19 +19,13 @@
 # Boston, MA 02111-1307, USA.
 #
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
 import copy
 import operator
 from itertools import chain
 from collections import OrderedDict
 
-from . import message
-
-from .message import Position
+from .sourcescanner import CTYPE_TYPEDEF, CSYMBOL_TYPE_TYPEDEF
+from .message import Position, warn
 from .utils import to_underscores
 
 
@@ -237,6 +231,9 @@ TYPE_DOUBLE = Type(target_fundamental='gdouble', ctype='gdouble')
 TYPE_LONG_DOUBLE = Type(target_fundamental='long double',
                         ctype='long double')
 TYPE_UNICHAR = Type(target_fundamental='gunichar', ctype='gunichar')
+# Platform-specific types
+TYPE_TIME_T = Type(target_fundamental='time_t', ctype='time_t')
+TYPE_OFF_T = Type(target_fundamental='off_t', ctype='off_t')
 
 # C types with semantics overlaid
 TYPE_GTYPE = Type(target_fundamental='GType', ctype='GType')
@@ -249,7 +246,7 @@ BASIC_TYPES = [TYPE_BOOLEAN, TYPE_INT8, TYPE_UINT8, TYPE_INT16,
                TYPE_UINT16, TYPE_INT32, TYPE_UINT32, TYPE_INT64,
                TYPE_UINT64, TYPE_CHAR, TYPE_SHORT, TYPE_USHORT, TYPE_INT,
                TYPE_UINT, TYPE_LONG, TYPE_ULONG, TYPE_SIZE, TYPE_SSIZE,
-               TYPE_LONG_LONG, TYPE_LONG_ULONG,
+               TYPE_LONG_LONG, TYPE_LONG_ULONG, TYPE_TIME_T, TYPE_OFF_T,
                TYPE_FLOAT, TYPE_DOUBLE,
                TYPE_LONG_DOUBLE, TYPE_UNICHAR, TYPE_GTYPE]
 
@@ -348,8 +345,10 @@ type_names['FILE*'] = TYPE_ANY
 # integers it's easy enough to expand them.
 type_names['size_t'] = type_names['gsize']
 type_names['ssize_t'] = type_names['gssize']
-type_names['time_t'] = TYPE_LONG
-type_names['off_t'] = type_names['gsize']
+type_names['uintptr_t'] = type_names['guintptr']
+type_names['intptr_t'] = type_names['gintptr']
+type_names['time_t'] = TYPE_TIME_T
+type_names['off_t'] = TYPE_OFF_T
 type_names['pid_t'] = TYPE_INT
 type_names['uid_t'] = TYPE_UINT
 type_names['gid_t'] = TYPE_UINT
@@ -367,6 +366,7 @@ PARAM_DIRECTION_INOUT = 'inout'
 PARAM_SCOPE_CALL = 'call'
 PARAM_SCOPE_ASYNC = 'async'
 PARAM_SCOPE_NOTIFIED = 'notified'
+PARAM_SCOPE_FOREVER = 'forever'
 
 PARAM_TRANSFER_NONE = 'none'
 PARAM_TRANSFER_CONTAINER = 'container'
@@ -565,6 +565,7 @@ properties."""
         self.deprecated = None
         self.deprecated_doc = None
         self.doc = None
+        self.doc_position = None
 
 
 class Node(Annotated):
@@ -632,9 +633,23 @@ GIName.  It's possible for nodes to contain or point to other nodes."""
     def add_file_position(self, position):
         self.file_positions.add(position)
 
+    def get_main_position(self):
+        if not self.file_positions:
+            return None
+
+        res = None
+        for position in self.file_positions:
+            if position.is_typedef:
+                res = position
+            else:
+                return position
+
+        return res
+
     def add_symbol_reference(self, symbol):
         if symbol.source_filename:
-            self.add_file_position(Position(symbol.source_filename, symbol.line))
+            self.add_file_position(Position(symbol.source_filename, symbol.line,
+                is_typedef=symbol.type in (CTYPE_TYPEDEF, CSYMBOL_TYPE_TYPEDEF)))
 
     def walk(self, callback, chain):
         res = callback(self, chain)
@@ -649,6 +664,11 @@ GIName.  It's possible for nodes to contain or point to other nodes."""
         pass
 
 
+class DocSection(Node):
+    def __init__(self, name=None):
+        Node.__init__(self, name)
+
+
 class Registered:
     """A node that (possibly) has gtype_name and get_type."""
     def __init__(self, gtype_name, get_type):
@@ -667,6 +687,9 @@ class Callable(Node):
         self.throws = not not throws
         self.instance_parameter = None  # Parameter
         self.parent = None  # A Class or Interface
+        self.finish_func = None
+        self.sync_func = None
+        self.async_func = None
 
     def _get_retval(self):
         return self._retval
@@ -717,6 +740,14 @@ class Callable(Node):
         raise ValueError("Unknown argument %s" % (name, ))
 
 
+class FunctionMacro(Node):
+    def __init__(self, name, parameters, symbol):
+        Node.__init__(self, name)
+        self.symbol = symbol
+        self.parameters = parameters
+        self.introspectable = False
+
+
 class Function(Callable):
 
     def __init__(self, name, retval, parameters, throws, symbol):
@@ -728,6 +759,9 @@ class Function(Callable):
         self.shadows = None             # C symbol string
         self.moved_to = None            # namespaced function name string
         self.internal_skipped = False   # if True, this func will not be written to GIR
+        self.set_property = None        # Property name
+        self.get_property = None        # Property name
+        self.is_inline = False
 
     def clone(self):
         clone = copy.copy(self)
@@ -750,7 +784,7 @@ class Function(Callable):
         # Returns GType
         rettype = self.retval.type
         if (not rettype.is_equiv(TYPE_GTYPE) and rettype.target_giname != 'Gtk.Type'):
-            message.warn("function '%s' returns '%r', not a GType" % (self.name, rettype))
+            warn("function '%s' returns '%r', not a GType" % (self.name, rettype))
             return False
 
         return True
@@ -857,7 +891,7 @@ class TypeContainer(Annotated):
         self.direction = direction
         if transfer is not None:
             self.transfer = transfer
-        elif typenode.is_const:
+        elif typenode and typenode.is_const:
             self.transfer = PARAM_TRANSFER_NONE
         else:
             self.transfer = None
@@ -948,12 +982,13 @@ class Bitfield(Node, Registered):
 
 class Member(Annotated):
 
-    def __init__(self, name, value, symbol, nick):
+    def __init__(self, name, value, symbol, nick=None, dump_name=None):
         Annotated.__init__(self)
         self.name = name
         self.value = value
         self.symbol = symbol
         self.nick = nick
+        self.dump_name = dump_name
         self.parent = None
 
     def _compare(self, other, op):
@@ -991,6 +1026,8 @@ class Compound(Node, Registered):
                  get_type=None,
                  c_symbol_prefix=None,
                  disguised=False,
+                 opaque=False,
+                 pointer=False,
                  tag_name=None):
         Node.__init__(self, name)
         Registered.__init__(self, gtype_name, get_type)
@@ -1000,6 +1037,8 @@ class Compound(Node, Registered):
         self.fields = []
         self.constructors = []
         self.disguised = disguised
+        self.opaque = opaque
+        self.pointer = pointer
         self.gtype_name = gtype_name
         self.get_type = get_type
         self.c_symbol_prefix = c_symbol_prefix
@@ -1086,6 +1125,8 @@ class Record(Compound):
                  get_type=None,
                  c_symbol_prefix=None,
                  disguised=False,
+                 opaque=False,
+                 pointer=False,
                  tag_name=None):
         Compound.__init__(self, name,
                           ctype=ctype,
@@ -1093,10 +1134,18 @@ class Record(Compound):
                           get_type=get_type,
                           c_symbol_prefix=c_symbol_prefix,
                           disguised=disguised,
+                          opaque=opaque,
+                          pointer=pointer,
                           tag_name=tag_name)
         # If non-None, this record defines the FooClass C structure
         # for some Foo GObject (or similar for GInterface)
         self.is_gtype_struct_for = None
+        # If non-None, this record has a copy function for heap
+        # allocated instances
+        self.copy_func = None
+        # If non-None, this record has a free function for heap
+        # allocated instances
+        self.free_func = None
 
 
 class Union(Compound):
@@ -1107,6 +1156,8 @@ class Union(Compound):
                  get_type=None,
                  c_symbol_prefix=None,
                  disguised=False,
+                 opaque=False,
+                 pointer=False,
                  tag_name=None):
         Compound.__init__(self, name,
                           ctype=ctype,
@@ -1114,7 +1165,15 @@ class Union(Compound):
                           get_type=get_type,
                           c_symbol_prefix=c_symbol_prefix,
                           disguised=disguised,
+                          opaque=opaque,
+                          pointer=pointer,
                           tag_name=tag_name)
+        # If non-None, this union has a copy function for heap
+        # allocated instances
+        self.copy_func = None
+        # If non-None, this union has a free function for heap
+        # allocated instances
+        self.free_func = None
 
 
 class Boxed(Node, Registered):
@@ -1154,6 +1213,7 @@ class Signal(Callable):
         self.detailed = detailed
         self.action = action
         self.no_hooks = no_hooks
+        self.emitter = None
 
 
 class Class(Node, Registered):
@@ -1163,7 +1223,8 @@ class Class(Node, Registered):
                  gtype_name=None,
                  get_type=None,
                  c_symbol_prefix=None,
-                 is_abstract=False):
+                 is_abstract=False,
+                 is_final=False):
         Node.__init__(self, name)
         Registered.__init__(self, gtype_name, get_type)
         self.ctype = ctype
@@ -1180,6 +1241,7 @@ class Class(Node, Registered):
         self.parent_chain = []
         self.glib_type_struct = None
         self.is_abstract = is_abstract
+        self.is_final = is_final
         self.methods = []
         self.virtual_methods = []
         self.static_methods = []
@@ -1269,6 +1331,9 @@ class Property(Node):
             self.transfer = PARAM_TRANSFER_NONE
         else:
             self.transfer = transfer
+        self.setter = None
+        self.getter = None
+        self.default_value = None
         self.parent = None  # A Class or Interface
 
 
